#!/usr/bin/env Rscript
# kinPlot.R
#
# Plot treekin output (or other kinetics)
# May 2015 -- Feb 2016
#
# Copyright 2015-2016 Felix Kuehnl, felix[at]bioinf.uni-leipzig.de
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################


library(RColorBrewer, quietly=TRUE)
library(argparser,    quietly=TRUE)


# Parse arguments
parser <- arg_parser('args')
parser <- add_argument( 
    parser, arg = '--ifile', nargs = Inf,
    help = paste(
        "Name(s) of input file(s). Separate multiple names by whitespace.",
        "Expects a table of float values where the first column represents",
        "a continous time scale and any sucessive column represents the",
        "population of a certain state at that time value. For naming of the",
        "states, use a header line, cf. option -H."
    )
)
parser <- add_argument( parser, arg = '--probMin', type = 'double', default = 0.03,
    help = "Minimal probability a state needs to acquire during the simulation to be visualized.")
parser <- add_argument(
    parser, arg = '-H', flag=TRUE,
    help = "Input file has a header which is used to name the state columns."
)
parser <- add_argument( parser, arg = '-l', flag=TRUE,
                        help = "Use linear instead of logarithmic time")
# parser <- add_argument( parser, arg = '--ofile', nargs=1,
#                        help = "Explicitly specify output file")
parser <- add_argument( parser, arg = '-c', flag=TRUE,
                        help = "Cut off plot when equilibrium is reached")
parser <- add_argument( 
    parser, arg = '--title', 
    help = paste( "Set title for plot. Defaults to output file name. Pass",
                  "empty string to disable title plotting.")
)
parser <- add_argument( 
    parser, arg = '--second-color-index', short = "-S", type = "numeric",
    help = paste( "Starting with the given index, a second gradient color",
                  "is used when plotting more than 12 curves."
           )
)
parser <- add_argument( 
    parser, arg = '--sum-states', default = "",
    help = paste0( "Comma-separated list of state indices. The probability ",
                   "of the selected states is summed up and added to the ",
                   "as an additional dashed curve. Use 'i-j' for continuous",
                   "indices. Also supports a leading label followed by ':'.",
                   "The first state has index 1 (time is not a state).")
)

dbg.args <- strsplit(
    "-H -c -i RS3.subopt_e25_T37.bar_minh1.nd.all.rates_lig1e-6.kin",
    split = " " 
)[[1]]

args <- parse_args( 
    parser 
#   , argv = dbg.args        # TODO disable, for debug only
)



prob.min           <- args$probMin
use.header         <- args$H
do.cutoff          <- args$c
linear.time        <- args$l            # logarithmic when false
input.files        <- args$ifile
sum.states.string  <- args$sum_states
second.color.index <- args$second_color_index
plot.title         <- args$title

# Use ColorBrewer categorical data colors for 3-12 curves
categorical.color.count <- 12           # for more curves, use gradient

if ( length(input.files) && is.na(input.files) )
    stop( "Specify at least one input file")


####################### Debugging code ###################
# work.dir <- "/homes/brauerei2/felix/scratch/examples/designed_RS_Wachsmuth13/dRS1/"
# 
# setwd( work.dir)
# args <- data.frame(
#     ifile   = "RS1.subopt_e25_T37.bar_minh1.nd.all.kin.head",
#     probMin = 0.03,
#     H       = TRUE,           # input has header
#     l       = FALSE,           # TRUE = linear, FALSE = logarithmic
#     ofile   = NA,
#     sum_states = 'testLabel:1-10,15-20'     
#
#     stringsAsFactors = FALSE
#   )
#paste( 'ifile:', args$ifile, '  prob.min:', args$probMin, '  -H: ', args$H,
#       '  -l: ', args$l, '  -o: ', args$ofile)
################# End of debugging code ##################


#################
##  Functions  ##
#################

# Parse string passed with option --sum-states, cf. help.
parse.states.string <- function( string) {
    # Process leading label
    split <- strsplit( string, split = ":", fixed = TRUE)[[1]]
    n <- length( split)
    label <- NA
    if(n>2) {
        stop( "Multiple labels / colons supplied in sum states string")
    }
    else if (n==2) {
        label  <- split[1]
        string <- split[2]
    }    
    
    # Separate groups or single indices
    split <- strsplit( string, split = ",", fixed = TRUE)[[1]]
    indices <- vector( mode = 'numeric')
    for (group in split) {
        group.split <- strsplit( group, split = "-", fixed = TRUE)[[1]]
        n <- length( group.split )
        if(n>2) {
            stop( paste0("Multiple hyphens supplied in sum states group: ",
                         group, "\n")
            )
        }
        else if (n==2) { # index range
            i <- as.numeric( group.split[1] )
            j <- as.numeric( group.split[2] )
            if (i>j) { 
                stop(paste0("Leading first index of sum states is greater ",
                            "than second one: ", group, "\n")
                )
            }
            indices <- c( indices, seq( from = i, to = j, by = 1))
        }
        else { # n==1, group is single index
            indices <- c( indices, as.numeric(group))
        }        
    }
    
    return( 
        list( 
            label = label, 
            indices = unique(sort(indices))        
        )                                
    )
}

# TODO make this better, do not consider only first column but all columns
# simply sum abs differences fore each column, take max
get.cutoff.time.index <- function( m, linear.time ) {
    final <- m[nrow(m),-1]   # final observations for each state
    # Substract final value of each column from all values in this column
    diff <- sweep( m[,-1], MARGIN = 2, STATS = final,
                   FUN = function(m, v) { abs(m-v) })
    max.diff = apply( diff, MARGIN = 1, max)  # get max diff in any state
    
    
    tmax <- tail(which( max.diff > 0.001), n = 1) # concentration of state 1 stable
    rm(diff, max.diff) # clean up, need memory
    if( length(tmax)==0) tmax <- 1 # concentration of state 1 remains constant   
    
    # Stretch by a factor to better fit legend into plot
    add.factor <- 1.18      # time to add after reaching equilibrium [1.2]
    start.time = m[1,1]
    time <- m[tmax,1]
    if ( linear.time ) {        # linear
        time <- start.time + (time - start.time) * add.factor      
    }
    else {                 # logarithmic
        #time <- max(time^add.factor, 1) # if t<1 exponentiation makes it smaller
        time <- start.time * (time/start.time)^add.factor
    }
    
    last.time  <- m[nrow(m),1]      # highest time value
    time <- min ( time, last.time)  # ensure we're not beyond simulated values
    time.index <- tail( which( m[,1] <= time), n = 1)                          
    # Keep at least two rows to be able to draw a line
    if ( time.index == 1 && nrow(m) >= 2) { time.index <- 2 }
    return( time.index)
}

# Plot single data column (vector) into existing plot
plot.col <- function( time, column, color, dashed = FALSE) {    
    lines( time, column, col=color, lwd=3.5,
           lty = ifelse(dashed, "dashed", "solid")
    )
}

# Plot results of a kinetic simulation (e.g. performed by Treekin)
# m - matrix, 1st col: time, 2nd col: mfe, other structs...
# chosen: vector of the columns of m that are supposed to be plotted
#         
plot.kin <- function(m, chosen, colors,
                     sum.states = list( label = NA, indices = c() ),
                     cutoff.time.index = NA, 
                     linear.time = FALSE,
                     plot.title = ""                     
            )
{
    if( length( m[,1]) < 2)
        stop( "Supplied matrix has less than 2 rows")
    
    # Scale all row sums 1 (particle number changes on dimer formation!)
    rsum = sum(m[1,-1])  
    if( abs(rsum-1) > 0.001 && rsum > 0.001)
    { 
        cat( 'Scaling matrix such that rows sum to 1...\n')
        rsum <- apply( X=m, MARGIN = 1, FUN = function(x) sum(x[-1]))
        m[,-1] = m[,-1]/rsum
    }
    rm(rsum)
    
    # cut time when mfe structure becomes stable
    
    if( !is.na(cutoff.time.index)) { # do cut-off?
        cutoff <- min(cutoff.time.index, nrow(m)) # prevent index out of bound
        if( cutoff != cutoff.time.index){
            cat( "WARNING: Time columns of input file differ!\n")
        }
        m <- m[1:cutoff, ,drop = FALSE]          
    }
    
    t <- m[,1] # Time col
    
    # Handle start times of t=0 for logarithmic scaling
    if( t[1] == 0  &&  linear.time == FALSE)
        t[1] = ifelse( length(t)>1, t[2]/10.,1)
    
    # Get indices of cols to plot 
    # chosen      <- which(apply( m[,-1], MARGIN=2, max) >= prob.min) 
    chosen.cols <- m[,chosen+1,drop=FALSE]
    col.names   <- colnames(m)[chosen+1]
    
    # Sum over sum states (option --sum-states)
    sum.col <- c()
    if( length(sum.states$indices) > 0 ) {
        sum.col <- apply( m[ , sum.states$indices+1, drop = FALSE ],
                          MARGIN = 1,
                          sum
                   )    
    }
    
    rm(m) # free memory before plotting        
    
    # Plotting    
    top.margin <- ifelse( plot.title == "", 0.2, 2 )
    par( mar = c(4, 4, top.margin, 0.5) + 0.1,
         bg = ifelse( length(chosen) > categorical.color.count,
                      "gray90",
                      "white"
              )
         )  

    t.start <- t[1]
    t.end   <- t[length(t)]

    plot(                                       # Initialize empty plot
        NULL,
        log  = ifelse(linear.time, '', 'x'),
        type = "l",
        main = ifelse( plot.title == "", NA, plot.title ),
        #xlim = c(t[1],t[length(t)]),
        xlim = c(t.start, t.end),
        ylim = c(0,1),
        xlab = ifelse(linear.time,'Time','log(Time)'),
        ylab = 'Macrostate density',
        xaxt = ifelse(linear.time, 's', 'n'),   # custom axis for log scale
    )

    if (!linear.time) {                         # custom axis for log scale
        exp.min    <-   floor(log10(t.start))
        exp.max    <- ceiling(log10(t.end  ))
        log.ticks  <-   outer(1:9, 10^(exp.min:exp.max))
        log.labels <-  ifelse(log10(log.ticks) %% 1 == 0, log.ticks, NA)
        axis(1, at=log.ticks, labels=log.labels, las=1)
    }
    
    for (i in 1:length(chosen)) {               # plotting chosen data columns
        plot.col(t, chosen.cols[,i], colors[i])
        #all[i] <- list(x=t,y=chosen.cols[,i]) 
    }
    if (length(sum.states$indices) > 0 ) {      # state sum column, option -s
        plot.col( t, sum.col, "black", dashed = TRUE)   
    }
    
    # Header given, use its names for labels/legend [m[,1] is the time] 
    if (use.header)       # remove the X that R adds in front of numbers  
        chosen <- sub( '^X', '', col.names) 
    if ( length(sum.col)>0 ) {
        label  <- ifelse( is.na(sum.states$label), "Sum", sum.states$label)
        chosen <- c(chosen, label)
        colors <- c(colors, "black")
    }
    legend(x = "right", legend = chosen, fill = colors, bg="lightgrey")
}


##################################
##  Gather cross-file settings  ##
##################################

cutoff.time.index <- NA   # plot until this time index
chosen            <- c()  # indices of columns to plot
for( input.file in input.files) {  
    
    # Read input matrix
    m <- as.matrix(
        read.table(
            file       = input.file,
            header     = use.header,
            colClasses = 'numeric',
        )
    )
    
    # Select columns to plot
    chosen <- union(
        chosen,
        which( apply( m[,-1], MARGIN=2, max) >= prob.min )
    )

    # Get cutoff time index
    cutoff.time.index <-
        max( cutoff.time.index,
             ifelse( do.cutoff,
                     get.cutoff.time.index(m, linear.time = linear.time),
                     length( m[,1])
             ),
             na.rm = TRUE
        )        
}

# Generate colors
n <- length(chosen)
# use ColornBrewer categorical data colors for 3-12 curves
if (n <= categorical.color.count)
{
    colors <- brewer.pal( max(n, 3), 'Paired') # package needs at least 3 :/
    if (n == 2) colors <- colors[c(1,3)]       # color 1 and 2 are too similar
} else {               # use a red-blue gradient
    if ( is.na(second.color.index) ) { # Use only one gradient
        #colorRamp <- colorRampPalette(c("blue","red"), space = "Lab")
        colorRamp <- colorRampPalette(c("blue","yellow"), space = "Lab")
        #colorRamp <- colorRampPalette(c("red","green"), space = "Lab")
        #colorRamp <- colorRampPalette(c("blue","white"), space = "Lab")        
        colors <- colorRamp(n)  
    } else {                           # Use two gradients 
        n1 <- length( which( chosen < second.color.index ) )
        n2 <- length(chosen) - n1    
        colorRamp1 <- colorRampPalette(c("blue","yellow"), space = "Lab")
        colorRamp2 <- colorRampPalette(c("red","green"),   space = "Lab")
        colors <- c( colorRamp1(n1), colorRamp2(n2))
    }
    
}    

# Parse sum states information, i.e. the column whose indices are returned are
# summed up and printed as a dashed curve.
sum.states <- parse.states.string( sum.states.string)

cat( "Plotting", n, "curves with", cutoff.time.index, "points each.\n" )


############################
##  Plotting input files  ##
############################

for( input.file in input.files) {      
    m <- as.matrix(
                    read.table(
                        file       = input.file,
                        header     = use.header,
                        colClasses = 'numeric',
                    )
         )

 
    out.file <- paste0( input.file, '.pdf' )
    pdf( out.file, title=out.file, width = 8, height = 4.5)
    plot.kin( m,
              chosen, 
              colors,
              sum.states,
              plot.title = ifelse( is.na(plot.title), out.file, args$title),
              cutoff.time.index = cutoff.time.index,
              linear.time       = linear.time
    )
    invisible(dev.off() )
    cat( "Output graph written to file '", out.file, "'\n", sep='' )
}
